#
# __init__.py - Admin interface initialization
#
# Copyright 2013, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>
#   Ilgiz Islamgulov <ilgizisl@gmail.com>
import logging

from flask import flash, request, redirect, url_for
from flask_admin.babel import gettext
import flask_admin
from blockerbugs import app, db
from blockerbugs.models.build import Build
from blockerbugs.models.release import Release
from blockerbugs.models.userinfo import UserInfo
from blockerbugs.models.milestone import Milestone
from blockerbugs.controllers.admin.build import BuildView
from blockerbugs.controllers.admin.auth import FasAuthModelView
from blockerbugs.controllers.users import check_admin_rights


class AdminIndexViewSqla(flask_admin.AdminIndexView):
    def _handle_view(self, name, **kwargs):
        return check_admin_rights()


admin = flask_admin.Admin(app, 'Blocker Bug Tracking Admin',
                        base_template='admin_layout.html',
                        template_mode='bootstrap3',
                        index_view=AdminIndexViewSqla())


class ReleaseView(FasAuthModelView):
    form_excluded_columns = ('last_update_sync', )

    def create_model(self, form):
        """
        Implement Release model constructor with non-default parameters support
        """
        try:
            model = self.model(form.number.data, form.active.data)
            self.session.add(model)
            self.on_model_change(form, model, True)
            self.session.commit()
        except Exception as ex:
            flash(gettext('Failed to create model. %(error)s', error=str(ex)),
                  'error')
            logging.exception('Failed to create model')
            self.session.rollback()
            return False
        return True


class MilestoneView(FasAuthModelView):
    form_excluded_columns = ('last_bug_sync', )

    def create_model(self, form):
        """
        Implement Milestone model constructor with non-default parameters support
        """
        try:
            model = self.model(form.release.data, form.version.data,
                               form.blocker_tracker.data, form.fe_tracker.data,
                               form.name.data, form.active.data,
                               form.current.data)
            #now only one milestone can succeeds (like a chain)
            if form.succeeded_by.data:
                model.succeeded_by = form.succeeded_by.data
            self.session.add(model)
            self.on_model_change(form, model, True)
            self.session.commit()
        except Exception as ex:
            flash(gettext('Failed to create model. %(error)s', error=str(ex)),
                  'error')
            logging.exception('Failed to create model')
            self.session.rollback()
            return False
        return True

class UserInfoView(FasAuthModelView):
    column_exclude_list = form_excluded_columns = ('api_key_hash', )

    edit_template = 'admin/edit_userinfo.html'

    def on_model_change(self, form, model, is_created):
        if not model.bz_user:
            model.bz_user = None


    @flask_admin.expose('/generate_api_key/', methods=('GET', ))
    def generate_api_key(self):
        """
            Generate user api key.
        """
        return_url = request.args.get('url') or url_for('.index_view')

        if not self.can_edit:
            return redirect(return_url)

        id = request.args.get('id')
        if id is None:
            return redirect(return_url)

        model = self.get_one(id)

        if model is None:
            return redirect(return_url)

        try:
            api_key = model.generate_api_key()
            self.session.add(model)
            self.session.commit()
        except Exception as ex:
            flash(gettext('Failed to generate API key. %(error)s', error=str(ex)),
                  'error')
            logging.exception('Failed to generate api key')
            self.session.rollback()
            return False

        return self.render('admin/generate_api_key.html',
                           client_id=model.fas_login,
                           api_key=api_key,
                           return_url=return_url)


admin.add_view(MilestoneView(Milestone, db.session))
admin.add_view(ReleaseView(Release, db.session))
admin.add_view(UserInfoView(UserInfo, db.session))
admin.add_view(BuildView(Build, db.session))
