#
# build.py - Build view for admin interface
#
# Copyright 2013, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>
#   Ilgiz Islamgulov <ilgizisl@gmail.com>

import koji
from flask import flash, request, url_for, redirect
import flask_wtf
import wtforms
from flask_admin import expose
from flask_admin.babel import gettext

from blockerbugs.controllers.admin.auth import FasAuthModelView
from blockerbugs.models.build import Build

class ReadOnlyTextField(wtforms.TextField):
    def __init__(self, *args, **kwargs):
        self.disabled = kwargs.pop('disabled', True)
        super(ReadOnlyTextField, self).__init__(*args, **kwargs)

    def __call__(self, **kwargs):
        if self.disabled:
            kwargs.setdefault('readonly', True)
        return self.widget(self, **kwargs)


class BuildForm(flask_wtf.FlaskForm):
    nvr = ReadOnlyTextField('NVR', [wtforms.validators.required()])

class BuildView(FasAuthModelView):
    column_searchable_list = (Build.nvr, )
    column_exclude_list = ('nvr', )

    form = BuildForm

    @expose('/new/', methods=('GET', 'POST'))
    def create_view(self):
        """
        Create new build by NVR, check build exists in Koji.
        """
        return_url = request.args.get('url') or url_for('.index_view')

        if not self.can_create:
            return redirect(return_url)

        form = self.form()
        form.nvr.disabled = False

        if form.validate_on_submit():
            try:
                Build.from_nvr(form.nvr.data)
                if '_add_another' in request.form:
                    flash(
                        gettext('Model was successfully created.'))
                    return redirect(
                        url_for('.create_view', url=return_url))
                else:
                    return redirect(return_url)
            except koji.GenericError as ex:
                form.nvr.errors.append(ex.message)

        return self.render(self.create_template,
                           form=form,
                           form_widget_args=self.form_widget_args,
                           return_url=return_url)
