#
# api.py - RESTful API handlers
#
# Copyright 2013, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>
#   Ilgiz Islamgulov <ilgizisl@gmail.com>
#

from flask import Blueprint, request

from blockerbugs import app, db
from blockerbugs.models.milestone import Milestone
from blockerbugs.models.update import Update
from blockerbugs.models.release import Release
from blockerbugs.models.bug import Bug
from blockerbugs.util import pagure_bot
from . import errors
from .utils import get_or_404, JsonResponse, check_signature

api_v0 = Blueprint('api', __name__, url_prefix='/api/v0')

ACCEPTED_BUGTYPES = [
    'proposed_blocker', 'proposed_fe',
    'rejected_blocker', 'rejected_fe',
    'accepted_blocker', 'accepted_fe',
    'accepted_0day', 'accepted_prevrel',
    'prioritized'
]


@api_v0.errorhandler(errors.RestApiError)
def api_error_handler(e):
    return JsonResponse(e.to_dict(), e.http_status_code)


def get_update_info(update):
    update_simple_fields = ['title', 'url', 'karma', 'stable_karma', 'status',
                            'pending']
    update_data = dict(
        (attr, getattr(update, attr)) for attr in update_simple_fields)
    update_data['release'] = update.release.number
    update_data['milestones'] = [{'version': m.version,
                                  'release': m.release.number, }
                                 for m in update.milestones]
    update_data['bugs'] = [
        {'bugid': bug.bugid,
         'type': [tp for tp in ACCEPTED_BUGTYPES if getattr(bug, tp)]}
        for bug in update.bugs
    ]
    return update_data


def get_bug_info(bug):
    bug_simple_fields = ['bugid', 'url', 'summary', 'component', 'active']
    bug_info = dict((attr, getattr(bug, attr)) for attr in bug_simple_fields)
    bug_info['type'] = [tp for tp in ACCEPTED_BUGTYPES if getattr(bug, tp)]
    return bug_info

@api_v0.route('/milestones/<int:rel_num>/<milestone_version>/updates')
def list_updates(rel_num, milestone_version):
    release = get_or_404(Release, number=rel_num)
    milestone = get_or_404(Milestone, release=release,
                           version=milestone_version)
    m_alias = db.aliased(Milestone)
    updates = Update.query.join(m_alias, Update.milestones).filter(
        m_alias.id == milestone.id)
    if 'bugtype' in request.args:
        bugtype = request.args['bugtype']
        if bugtype in ACCEPTED_BUGTYPES:
            bug_attr = getattr(Bug, bugtype)
            updates = updates.filter(Update.bugs.any(bug_attr == True))
        else:
            raise errors.InvalidArgumentError(arg_name='bugtype')
    updates = updates.order_by(Update.date_submitted.desc()).all()
    updates_info = [get_update_info(up) for up in updates]
    return JsonResponse(updates_info)


@api_v0.route('/milestones/<int:rel_num>/<milestone_version>/bugs')
def list_bugs(rel_num, milestone_version):
    release = get_or_404(Release, number=rel_num)
    milestone = get_or_404(Milestone, release=release,
                           version=milestone_version)
    bugs = Bug.query.filter_by(milestone=milestone)
    if 'bugtype' in request.args:
        bugtype = request.args['bugtype']
        if bugtype in ACCEPTED_BUGTYPES:
            bug_attr = getattr(Bug, bugtype)
            bugs = bugs.filter(bug_attr == True)
        else:
            raise errors.InvalidArgumentError(arg_name='bugtype')
    bugs = bugs.order_by(Bug.component, Bug.bugid).all()
    bugs_info = [get_bug_info(bug) for bug in bugs]
    return JsonResponse(bugs_info)

@api_v0.route('/milestones/current')
def get_current_milestone():
    current_milestone = get_or_404(Milestone, current=True)
    return JsonResponse(current_milestone.simple())


@api_v0.route('/webhook', methods=['POST'])
def pagure_webhook():
    if not app.config['PAGURE_BOT_ENABLED']:
        msg = 'Pagure bot disabled, ignoring request'
        app.logger.info(msg)
        return JsonResponse({'msg': msg})

    if not check_signature(request.headers, request.data):
        msg = 'Wrong signature, ignoring.'
        app.logger.debug(msg)
        return JsonResponse({'msg': msg})

    data = request.json
    if data['topic'] not in ['issue.comment.added', 'issue.comment.edited']:
        msg = 'Ignoring message with topic %s' % data['topic']
        app.logger.debug(msg)
        return JsonResponse({'msg': msg})

    issue_id = data.get('msg', {}).get('issue', {}).get('id', {})
    if issue_id:
        pagure_bot.webhook_handler(issue_id)
        msg = 'Message successfully parsed'
        app.logger.debug(msg)
        return JsonResponse({'msg': msg})
    else:
        msg = 'Unable to parse issue id from received message'
        app.logger.debug(msg)
        return JsonResponse({'msg': msg})
