#
# utils.py - API utils
#
# Copyright 2013, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>
#   Ilgiz Islamgulov <ilgizisl@gmail.com>
#
import json
import datetime
from functools import wraps

import hmac
import hashlib

from flask import Response, request

from blockerbugs import app
from blockerbugs.models import EnumSymbol
from blockerbugs.models.userinfo import UserInfo
from .errors import MalformedJSONError, NoSuchObjectError, AuthFailedError


class JsonEncoder(json.JSONEncoder):
    """
    JSONEncoder subclass that knows how to encode date/time and DeclEnum.
    """

    def default(self, obj):
        if isinstance(obj, EnumSymbol):
            return obj.description
        if isinstance(obj, datetime.datetime):
            r = obj.isoformat()
            if obj.microsecond:
                r = r[:23] + r[26:]
            if r.endswith('+00:00'):
                r = r[:-6] + 'Z'
            return r
        elif isinstance(obj, datetime.date):
            return obj.isoformat()
        elif isinstance(obj, datetime.time):
            #determines if a given datetime.datetime is aware.
            if obj.tzinfo is not None and obj.tzinfo.utcoffset(
                    obj) is not None:
                raise ValueError("JSON can't represent timezone-aware times.")
            r = obj.isoformat()
            if obj.microsecond:
                r = r[:12]
            return r
        return super(JsonEncoder, self).default(obj)


class JsonResponse(Response):
    default_mimetype = 'application/json'

    def __init__(self, response=None, *args, **kwargs):
        if response is None:
            response = []
        data = json.dumps(response, cls=JsonEncoder)
        super(JsonResponse, self).__init__(data, *args, **kwargs)


def json_loads(data):
    try:
        return json.loads(data)
    except (ValueError, TypeError) as e:
        raise MalformedJSONError(detail=str(e))


def get_or_404(model, *args, **kwargs):
    rv = model.query.filter_by(*args, **kwargs).first()
    if rv is None:
        raise NoSuchObjectError(obj_type=model.__name__)
    return rv

def check_api_rights(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        headers = request.headers
        user_info = UserInfo.query.filter_by(fas_login=headers.get('X-Auth-User')).first()
        if user_info and user_info.check_api_key(headers.get('X-Auth-Key')):
            return f(*args, **kwargs)
        raise AuthFailedError()
    return decorated_function


def check_signature(headers, payload):
    key = bytes(app.config['PAGURE_REPO_WEBHOOK_KEY'], encoding="ascii")

    hashhex = hmac.new(key, payload, hashlib.sha1).hexdigest()

    return hashhex == headers.get('X-Pagure-Signature')
