#
# validators.py - Various data validators
#
# Copyright 2013, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>
#   Ilgiz Islamgulov <ilgizisl@gmail.com>
#
import re

from iso8601 import parse_date, ParseError
from .errors import ValidationError

try:
    basestring = basestring
except NameError:
    basestring = str

class BaseValidator(object):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.configure(args, kwargs)

    def configure(self, args, kwargs):
        self.required = kwargs.pop('required', True)
        self.default_name = kwargs.pop('default_name', 'Data')
        self._name = kwargs.pop('name', None)

    def _get_name(self):
        if self._name is not None:
            return self._name
        return self.default_name

    def _set_name(self, new_name):
        self._name = new_name

    name = property(_get_name, _set_name)

    def __call__(self, data=None):
        """
        Validate data
        @return True if data correct
        """
        self.raw_data = data
        self._validate()
        return True

    def _validate(self):
        if not self.required and self.raw_data is None:
            return True
        self._check_data()

    def _check_data(self):
        """
        Validation function that does not check required flag.
        """
        raise NotImplementedError('To use this class inherit from it '
            'and define _check_data method')


class NumericValidator(BaseValidator):
    numeric_type = None

    def configure(self, args, kwargs):
        self.max = kwargs.pop('max', None)
        self.min = kwargs.pop('min', None)
        super(NumericValidator, self).configure(args, kwargs)

    def _check_data(self):
        if not isinstance(self.raw_data, self.numeric_type):
            raise ValidationError('%s must be integer' % (self.name))
        if self.max is not None and self.raw_data > self.max:
            raise ValidationError('%s must be smaller than %i' % (self.name,
                                                                  self.max))
        if self.max is not None and self.raw_data < self.min:
            raise ValidationError('%s must be larger than %i' % (self.name,
                                                                 self.min))


class IntegerValidator(NumericValidator):
    numeric_type = int

class StringValidator(BaseValidator):
    def _check_data(self):
        if not isinstance(self.raw_data, basestring):
            raise ValidationError('%s must be string' % (self.name))


class BooleanValidator(BaseValidator):
    def _check_data(self):
        return bool(self.raw_data)


class NoneValidator(BaseValidator):
    def _check_data(self):
        return self.raw_data is None


class ConstValidator(BaseValidator):
    def configure(self, args, kwargs):
        self.const = args[0]
        super(ConstValidator, self).configure(args, kwargs)

    def _check_data(self):
        if self.const != self.raw_data:
            raise ValidationError('%s must be equal to %s' % (self.name,
                                                              str(self.const)))


class DictValidator(BaseValidator):
    def configure(self, args, kwargs):
        if not isinstance(args[0], dict):
            raise TypeError('Argument must be dict')
        self.items_validators = args[0]
        for key, validator in self.items_validators.items():
            if validator.name == validator.default_name:
                validator.name = key
        super(DictValidator, self).configure(args, kwargs)

    def _check_data(self):
        if not isinstance(self.raw_data, dict):
            raise ValidationError('%s must be dict' % (self.name))
        for key, validator in self.items_validators.items():
            validator(self.raw_data.get(key, None))


class ChoicesValidator(BaseValidator):
    def configure(self, args, kwargs):
        self.choices = set(args[0])
        super(ChoicesValidator, self).configure(args, kwargs)

    def _check_data(self):
        if not self.raw_data in self.choices:
            raise ValidationError(
                '%s must be one of %s' % (self.name, str(self.choices)))


class TypeValidator(BaseValidator):
    def configure(self, args, kwargs):
        self.type = args[0]
        super(TypeValidator, self).configure(args, kwargs)

    def _check_data(self):
        if not isinstance(self.raw_data, self.type):
            raise ValidationError(
                '%s must be instance of type %s' % (self.name,
                                                    self.type.__name__))


class ListValidator(BaseValidator):
    def configure(self, args, kwargs):
        self.values_type = args[0]
        super(ListValidator, self).configure(args, kwargs)

    def _check_data(self):
        for item in self.raw_data:
            if not isinstance(item, self.values_type):
                raise ValidationError(
                    '%s must be instance of type %s' %
                    (self.name, self.values_type.__name__))


class DateTimeValidator(BaseValidator):
    def _check_data(self):
        try:
            parse_date(self.raw_data)
        except ParseError:
            raise ValidationError('%s must be datetime in ISO 8601 format' % (self.name))


class StringRegexValidator(BaseValidator):
    def configure(self, args, kwargs):
        self.regex = re.compile(args[0])
        super(StringRegexValidator, self).configure(args, kwargs)

    def _check_data(self):
        if not self.regex.match(self.raw_data):
            raise ValidationError('%s invalid input string' % (self.name))


class UrlValidator(BaseValidator):
    url_regex = re.compile(
        r'^(?:http|ftp)s?://' # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|' #domain...
        r'localhost|' #localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})' # ...or ip
        r'(?::\d+)?' # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)

    def _check_data(self):
        if not self.url_regex.match(self.raw_data):
            raise ValidationError('%s must be valid URL' % (self.name))
