#
# main.py - interface for the main page
#
# Copyright 2012, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>

from flask import Blueprint, render_template, redirect, url_for, abort, g, flash, make_response, request
import datetime
from sqlalchemy import func, desc, or_, and_
import bugzilla
from flask_fas_openid import fas_login_required

from blockerbugs import app, db, __version__
from blockerbugs.util.bz_interface import BlockerProposal, BZInterfaceError
from blockerbugs.models.bug import Bug
from blockerbugs.models.milestone import Milestone
from blockerbugs.models.update import Update
from blockerbugs.models.release import Release
from blockerbugs.models.userinfo import UserInfo
from blockerbugs.controllers.forms import BugProposeForm, FasBugzillaForm

main = Blueprint('main', __name__)


@app.before_request
def before_request():
    g.milestones = Milestone.query.filter_by(active=True).order_by(Milestone.name).all()
    g.version = __version__


def get_recent_modifications(milestoneid):
    now = datetime.datetime.utcnow()
    modcutoff = now - datetime.timedelta(1)
    recentmods = Bug.query.filter_by(milestone_id=milestoneid).filter(
        Bug.last_bug_sync > modcutoff).all()
    return [bug.bugid for bug in recentmods]


def get_recent_whiteboard_change(milestoneid):
    now = datetime.datetime.utcnow()
    modcutoff = now - datetime.timedelta(1)
    recentmods = Bug.query.filter_by(milestone_id=milestoneid).filter(
        Bug.last_whiteboard_change > modcutoff).all()
    return [bug.bugid for bug in recentmods]


def get_milestone_bugs(milestone):
    bugz = {'Accepted Blockers':[], 'Accepted 0-day Blockers':[], 'Accepted Previous Release Blockers':[], 'Proposed Blockers':[], 'Accepted Freeze Exceptions':[], 'Proposed Freeze Exceptions':[], 'Prioritized Bugs':[]}

    bugz['Accepted Blockers'] = Bug.query.filter_by(milestone=milestone,
                                                    accepted_blocker=True,
                                                    active=True).order_by(Bug.component, Bug.bugid).all()
    bugz['Accepted 0-day Blockers'] = Bug.query.filter_by(milestone=milestone,
                                                    accepted_0day=True,
                                                    active=True).order_by(Bug.component, Bug.bugid).all()
    bugz['Accepted Previous Release Blockers'] = Bug.query.filter_by(milestone=milestone,
                                                    accepted_prevrel=True,
                                                    active=True).order_by(Bug.component, Bug.bugid).all()
    bugz['Proposed Blockers'] = Bug.query.filter_by(milestone=milestone,
                                                    proposed_blocker=True,
                                                    active=True).order_by(Bug.component, Bug.bugid).all()
    bugz['Accepted Freeze Exceptions'] = Bug.query.filter_by(milestone=milestone,
                                                    accepted_fe=True,
                                                    active=True).order_by(Bug.component, Bug.bugid).all()
    bugz['Proposed Freeze Exceptions'] = Bug.query.filter_by(milestone=milestone,
                                                    proposed_fe=True,
                                                    active=True).order_by(Bug.component, Bug.bugid).all()
    bugz['Prioritized Bugs'] = Bug.query.filter_by(milestone=milestone,
                                                    prioritized=True,
                                                    active=True).order_by(Bug.component, Bug.bugid).all()
    return bugz


def get_milestone_pending_stable_updates(milestone):
    """
    return list of updates associated with milestone,
    which are 'pending stable' in bodhi
    and are marked as fixing any non-rejected blocker/fe bug
    """
    m_alias = db.aliased(Milestone)
    updates = Update.query.join(m_alias, Update.milestones).filter(
        m_alias.id == milestone.id,
        Update.status == u'stable', Update.pending == True,
        Update.bugs.any(
            or_(Bug.accepted_fe == True, Bug.accepted_blocker == True,
                Bug.accepted_0day == True, Bug.accepted_prevrel == True,
                Bug.proposed_fe == True, Bug.proposed_blocker == True))).all()
    return updates


def get_milestone_updates_testing(milestone):
    """
    return list of updates associated with milestone,
    which are 'updates testing' in bodhi (not updates testing pending)
    and are marked as fixing any non-rejected blocker/fe bug
    """
    m_alias = db.aliased(Milestone)
    updates = Update.query.join(m_alias, Update.milestones).filter(
        m_alias.id == milestone.id, Update.status == 'testing',
        Update.karma < Update.stable_karma,
        or_(Update.pending != True, Update.pending == None),
        Update.bugs.any(
            or_(Bug.accepted_fe == True, Bug.accepted_blocker == True,
                Bug.accepted_0day == True, Bug.accepted_prevrel == True,
                Bug.proposed_fe == True, Bug.proposed_blocker == True))).all()
    return updates

def get_milestone_all_nonstable_blocker_fixes(milestone):
    """
    return list of all non-stable, non-obsolete updates which are
    marked as fixing any accepted blocker bug for milestone
    """
    m_alias = db.aliased(Milestone)
    updates = Update.query.join(m_alias, Update.milestones).filter(
        m_alias.id == milestone.id, ~Update.status.in_(['obsolete', 'deleted', 'unpushed']),
        or_(Update.status != 'stable', Update.pending == True),
        Update.bugs.any(
            and_(
                Bug.milestone_id == milestone.id,
                or_(
                    Bug.accepted_blocker == True, Bug.accepted_0day == True
                )
            )
        )
    ).all()
    return updates

def get_milestone_all_nonstable_fe_fixes(milestone):
    """
    return list of all non-stable, non-obsolete updates which are
    marked as fixing any accepted FE bug for milestone
    """
    m_alias = db.aliased(Milestone)
    updates = Update.query.join(m_alias, Update.milestones).filter(
        m_alias.id == milestone.id,  ~Update.status.in_(['obsolete', 'deleted', 'unpushed']),
        or_(Update.status != 'stable', Update.pending == True),
        Update.bugs.any(
            and_(
                Bug.milestone_id == milestone.id,
                Bug.accepted_fe == True
            )
        )
    ).all()
    return updates


def get_current_milestone():
    current_milestone = Milestone.query.filter_by(current=True).first()
    # if no current release is set, just grab one
    if not current_milestone:
        current_milestone = Milestone.query.first()
    return current_milestone


def get_milestone_info(milestone):
    return {'number': milestone.release.number,
            'phase': milestone.version.title(),
            'criteria_url': 'http://fedoraproject.org/wiki/Fedora_%d_%s_Release_Criteria' % (milestone.release.number, milestone.version.title()),
            'last_updated': milestone.last_bug_sync.strftime('%Y-%m-%d %H:%M:%S UTC') if milestone.last_bug_sync else None,
            'blocker_tracker': milestone.blocker_tracker,
            'fe_tracker': milestone.fe_tracker
    }


@main.route('/')
def index():
    if app.debug:
        app.logger.debug('rendering index')
    return render_template('index.html', title='Fedora Blocker Bugs')


@main.route('/<int:num>/<release_name>')
def old_display_buglist(num, release_name):
    return redirect(url_for('.display_buglist', num=num, release_name=release_name))


@main.route('/milestone/<int:num>/<release_name>')
def default_milestone_display(num, release_name):
    return redirect(url_for('.display_buglist', num=num, release_name=release_name))


@main.route('/milestone/<int:num>/<release_name>/buglist')
def display_buglist(num, release_name):
    release = Release.query.filter_by(number=num).first()
    milestone = Milestone.query.filter_by(release=release, version=release_name).first()
    if not milestone:
        abort(404)
    release_info = get_milestone_info(milestone)
    bugz = get_milestone_bugs(milestone)
    recent_bugs = get_recent_modifications(milestone.id)
    whiteboard_change = get_recent_whiteboard_change(milestone.id)
    return render_template('blocker_list.html',
                           buglists=bugz,
                           recent=recent_bugs,
                           wb_change=whiteboard_change,
                           info=release_info,
                           title="Fedora %s %s Blocker Bugs" % (
                           release_info['number'], release_info['phase']))


@main.route('/bug/<int:bugid>/updates')
def display_bug_updates(bugid):
    bug = Bug.query.filter_by(bugid=bugid).first()
    if not bug:
        abort(404)
    packagename = bug.component
    updates = bug.updates.filter(
        Update.status != 'obsolete',
        Update.status != 'deleted',
        Update.status != 'unpushed'
    ).all()
    return render_template('bug_tooltip.html', packagename=packagename, updates=updates)


@main.route('/milestone/<int:num>/<release_name>/irc')
def display_irc_blockers(num, release_name):
    release = Release.query.filter_by(number=num).first()
    milestone = Milestone.query.filter_by(release=release, version=release_name).first()
    if not milestone:
        abort(404)
    bugz = get_milestone_bugs(milestone)
    milestone_info = get_milestone_info(milestone)

    response = make_response(render_template('irc_format.txt', buglists=bugz, info=milestone_info))
    response.mimetype = 'text/plain'
    return response


@main.route('/milestone/<int:num>/<release_name>/updates')
def display_release_updates(num, release_name):
    release = Release.query.filter_by(number=num).first()
    milestone = Milestone.query.filter_by(release=release, version=release_name).first()
    if not milestone:
        abort(404)
    milestone_info = get_milestone_info(milestone)
    updates = {
        'Non Stable Updates': get_milestone_pending_stable_updates(milestone),
        'Updates Needing Testing': get_milestone_updates_testing(milestone),
    }
    return render_template('update_list.html',
                           info=milestone_info,
                           updates=updates,
                           title="Fedora %s %s Blocker Bug Updates" % (milestone_info['number'], milestone_info['phase']))

@main.route('/milestone/<int:num>/<release_name>/requests')
def display_release_requests(num, release_name):
    release = Release.query.filter_by(number=num).first()
    milestone = Milestone.query.filter_by(release=release, version=release_name).first()
    if not milestone:
        abort(404)
    milestone_info = get_milestone_info(milestone)
    blockers = get_milestone_all_nonstable_blocker_fixes(milestone)
    fes = get_milestone_all_nonstable_fe_fixes(milestone)
    # if an update fixes both blockers and fes, drop it from fe list
    fes = [fe for fe in fes if fe not in blockers]

    response = make_response(render_template('requests.txt', blockers=blockers, fes=fes, milestone=milestone.id))
    response.mimetype = 'text/plain'
    return response

@main.route('/milestone/<int:num>/<milestone_name>/need_testing')
def display_updates_need_testing(num, milestone_name):
    release = Release.query.filter_by(number=num).first()
    milestone = Milestone.query.filter_by(release=release, version=milestone_name).first()
    if not milestone:
        abort(404)
    milestone_info = get_milestone_info(milestone)
    updates = get_milestone_updates_testing(milestone)
    return render_template('update_need_testing.html', updates=updates,
                           info=milestone_info,
                           title="Fedora %s %s Blocker Bug Updates" % (milestone_info['number'], milestone_info['phase']))


@main.route('/milestone/<int:num>/<milestone_name>/info')
def display_milestone_info(num, milestone_name):
    release = Release.query.filter_by(number=num).first()
    milestone = Milestone.query.filter_by(release=release, version=milestone_name).first()
    if not milestone:
        abort(404)
    milestone_info = get_milestone_info(milestone)
    return render_template('milestone_info.html', info=milestone_info,
                           title="Fedora %s %s Blocker Bug Updates" % (milestone_info['number'], milestone_info['phase']),
                           bz_url=app.config['BUGZILLA_URL'])


@main.route('/milestone/<int:num>/<milestone_name>/stats')
def display_stats(num, milestone_name):
    release = Release.query.filter_by(number=num).first()
    milestone = Milestone.query.filter_by(release=release, version=milestone_name).first()
    if not milestone:
        abort(404)
    milestone_info = get_milestone_info(milestone)

    blockercount = {}
    blockercount['accepted'] = Bug.query.filter(Bug.milestone == milestone,
                                                or_(Bug.accepted_blocker == True,
                                                    Bug.accepted_0day == True,
                                                    Bug.accepted_prevrel == True),
                                                Bug.active == True).count()

    blockercount['proposed'] = Bug.query.filter_by(milestone=milestone,
                                                   proposed_blocker=True,
                                                   active=True).count()

    fecount = {}
    fecount['accepted'] = Bug.query.filter_by(milestone=milestone,
                                              accepted_fe=True,
                                              active=True).count()
    fecount['proposed'] = Bug.query.filter_by(milestone=milestone,
                                              proposed_fe=True,
                                              active=True).count()

    bugcount = {}
    bugcount['fe'] = Bug.query.filter(Bug.milestone == milestone,
                                        or_(Bug.accepted_fe == True, Bug.proposed_fe == True),
                                        Bug.active == True).count()
    bugcount['blocker'] = Bug.query.filter(Bug.milestone == milestone,
                                        or_(Bug.accepted_blocker == True, Bug.accepted_0day == True,
                                            Bug.accepted_prevrel == True, Bug.proposed_blocker == True),
                                        Bug.active == True).count()

    blocker_components = db.session.query(func.count(Bug.component),
                            Bug.component).filter(Bug.milestone==milestone,
                            or_(Bug.accepted_blocker == True, Bug.accepted_0day == True,
                                Bug.accepted_prevrel == True, Bug.proposed_blocker == True),
                            Bug.active==True).group_by(Bug.component).order_by(desc(func.count(Bug.component))).all()

    fe_components = db.session.query(func.count(Bug.component),
                        Bug.component).filter(Bug.milestone==milestone,
                        or_(Bug.accepted_fe == True, Bug.proposed_fe == True),
                        Bug.active==True).group_by(Bug.component).order_by(desc(func.count(Bug.component))).all()

    return render_template('milestone_stats.html', info=milestone_info,
                           blockercount=blockercount,
                           fecount=fecount,
                           bugcount=bugcount,
                           blockercomponents=blocker_components,
                           fecomponents=fe_components,
                           title="Fedora %s %s Stats" % (milestone_info['number'], milestone_info['phase']))


@main.route('/current')
def display_current():
    current_milestone = get_current_milestone()
    return redirect(url_for('.display_buglist', num=current_milestone.release.number, release_name=current_milestone.version))


@main.route('/current/irc')
def display_current_irc():
    current_milestone = get_current_milestone()
    return redirect(url_for('.display_irc_blockers', num=current_milestone.release.number, release_name=current_milestone.version))


def get_bugzilla():
    # create bz interface object using the app's config settings
    if not app.testing:
        if app.config.get('FAS_USER'):

            app.logger.debug('logging into bugzilla (%s) as %s' % (app.config['BUGZILLA_XMLRPC'], app.config['FAS_USER']))

            return bugzilla.RHBugzilla4(url=app.config['BUGZILLA_XMLRPC'],
                                        user=app.config['FAS_USER'],
                                        password=app.config['FAS_PASSWORD'],
                                        cookiefile=None,
                                        tokenfile=None)
        else:
            app.logger.warning('No FAS_USER configured. bugzilla modifications will NOT work!')
            return bugzilla.RHBugzilla4(url=app.config['BUGZILLA_XMLRPC'], cookiefile=None, tokenfile=None)
    else:
        app.logger.error('Bugzilla connections with test configuration is not supported!')
        return None


@main.route('/fas_bugzilla', methods=['GET', 'POST'])
@fas_login_required
def fas_bugzilla():
    fas_bz_form = FasBugzillaForm()

    if fas_bz_form.validate_on_submit():
        try:
            # try logging into bz if user/pass are valid
            bz = bugzilla.RHBugzilla4(url=app.config['BUGZILLA_XMLRPC'],
                                        user=fas_bz_form.bz_user.data,
                                        password=fas_bz_form.bz_pass.data,
                                        cookiefile=None, tokenfile=None)
            bz.disconnect()

            #create new user_info if user with given FAS-login does not exist
            #if user_with given FAS-login exists just update user_info.bz_user
            user_info = UserInfo.query.filter_by(fas_login=g.fas_user.username).first()
            if user_info:
                user_info.bz_user = fas_bz_form.bz_user.data
            else:
                user_info = UserInfo(g.fas_user.username, fas_bz_form.bz_user.data)
            db.session.add(user_info)
            db.session.commit()
            return redirect(url_for('.propose_bug'))
        except bugzilla.BugzillaError as e:
            app.logger.info(e)
            flash(e)

    return render_template('fas_bugzilla.html', title='Link your FAS and Red Hat Bugzilla Accounts',
                            fas_bz_form=fas_bz_form)


@main.route('/propose_bug', methods=['GET', 'POST'])
@fas_login_required
def propose_bug():
    user_info = UserInfo.query.filter_by(fas_login=g.fas_user.username).first()
    if not user_info or not user_info.bz_user:
        return redirect(url_for('.fas_bugzilla'))

    current_milestone = get_current_milestone()
    bugform = BugProposeForm()
    bugform.milestone.choices = [(m.id, m.name) for m in Milestone.query.filter_by(active=True).order_by(Milestone.id).all()]
    bugform.milestone.default = current_milestone.id

    # we have to process the form after setting the milestone default so that it is properly set
    # by passing in the request.form, anything submitted as part of a POST will be inserted as form data
    bugform.process(request.form, bz_user = user_info.bz_user)

    if bugform.validate_on_submit():
        app.logger.debug('bugid: %i' % bugform.bugid.data)
        app.logger.debug('bz_user: %s' % user_info.bz_user)
        app.logger.debug('milestone: %i' % bugform.milestone.data)
        app.logger.debug('blocker: %s' % bugform.blocker.data)
        app.logger.debug('freeze_exception: %s' % bugform.freeze_exception.data)
        app.logger.debug('justification: %s' % bugform.justification.data)

        selected_milestone = Milestone.query.filter_by(id=bugform.milestone.data).first()
        bugid = bugform.bugid.data

        trackers = {'blocker': selected_milestone.blocker_tracker,
                    'fe': selected_milestone.fe_tracker}

        app.logger.debug('selected trackers: %s' % trackers)

        # The form does basic datatype validation but we need to check for valid
        # data
        bz = get_bugzilla()
        proposer = BlockerProposal(bz, bugid, trackers, bugform.blocker.data,
                                   bugform.freeze_exception.data, user_info.bz_user)

        # check bug to make sure it exists and that it isn't closed
        try:
            proposer.check_proposed_bug()
        except BZInterfaceError as e:
            bugform.bugid.errors = [e.msg]

        # check to make sure that any proposals aren't being re-done
        if bugform.blocker.data:
            app.logger.debug('checking for valid blocker proposal')
            if not proposer.check_blocker_proposal():
                bugform.blocker.errors = ['Bug %i is already proposed as a blocker' % bugid]
        if bugform.freeze_exception.data:
            app.logger.debug('checking for valid freeze exception proposal')
            if not proposer.check_fe_proposal():
                bugform.freeze_exception.errors = ['Bug %i is already proposed as a freeze exception' % bugid]

        if len(bugform.errors) == 0:
            user = 'Fedora user %s' % g.fas_user.username
            app.logger.info('%s proposing %i as %s for %s' % (
                user, bugid, proposer.get_tracker_type(), selected_milestone.name))
            try:
                proposer.propose_bugs(user, selected_milestone.name, bugform.justification.data)
                return render_template('thanks.html', bugid=bugid,
                                       isblocker=bugform.blocker.data,
                                       isfe=bugform.freeze_exception.data)
            except BZInterfaceError as e:
                bugform.bugid.errors = [e.msg]
                app.logger.info('bug proposal form errors: %s' % e.msg)
        else:
            app.logger.info('bug proposal form errors: %s' % bugform.errors)

    # this assumes that the current milestone is the release against which the
    # blocker is being proposed. It isn't an ideal assumption but it should hold
    # for most use cases.

    current_release = current_milestone.release.number
    return render_template('propose_bug.html', title='Propose a Blocker or Freeze Exception',
                            bugform=bugform, release=current_release)
