#
# update.py - database model for bodhi updates
#
# Copyright 2012, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>

from blockerbugs import db
from blockerbugs.models.bug import Bug

# unused imports, but required by dynamic relationships:
from blockerbugs.models.bug import Bug
from blockerbugs.models.release import Release
from blockerbugs.models.milestone import Milestone


update_fixes = db.Table(
    'update_fixes',
    db.Column('bug_id', db.Integer, db.ForeignKey('bug.id')),
    db.Column('update_id', db.Integer, db.ForeignKey('update.id'))
)

update_milestones = db.Table(
    'update_milestones',
    db.Column('milestone_id', db.Integer, db.ForeignKey('milestone.id')),
    db.Column('update_id', db.Integer, db.ForeignKey('update.id'))
)


class Update(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    title = db.Column(db.Text, unique=False)
    url = db.Column(db.Text, unique=False)
    karma = db.Column(db.Integer, unique=False)
    stable_karma = db.Column(db.Integer, unique=False)
    status = db.Column(db.String(80), unique=False)
    request = db.Column(db.String(80), unique=False, nullable=True)
    pending = db.Column(db.Boolean, unique=False)
    date_submitted = db.Column(db.DateTime)
    date_pushed_testing = db.Column(db.DateTime, nullable=True)
    date_pushed_stable = db.Column(db.DateTime, nullable=True)
    date_obsoleted = db.Column(db.DateTime, nullable=True)
    bugs = db.relationship('Bug',
                           secondary=update_fixes,
                           backref=db.backref('updates',
                                              lazy='dynamic',
                                              order_by=(status.desc(), pending.desc())))
    release_id = db.Column(db.Integer, db.ForeignKey('release.id'))
    release = db.relationship('Release',
                              backref=db.backref('bugs', lazy='dynamic'))
    milestones = db.relationship('Milestone', secondary=update_milestones,
                                 backref='updates')

    def __init__(self, title, url, karma, status, bugs, release,
                 milestones, stable_karma=3, date_submitted=None):
        self.title = title
        self.url = url
        self.karma = karma
        self.status = status
        self.request = None
        self.bugs = bugs
        self.date_submitted = date_submitted
        self.release = release
        if milestones:
            self.milestones = milestones
        self.date_pushed_testing = None
        self.date_pushed_stable = None
        self.date_obsoleted = None

    def __str__(self):
        return 'update: %s' % (self.title)

    def sync(self, updateinfo, fullsync=False):
        self.title = updateinfo['title']
        self.url = updateinfo['url']
        self.karma = updateinfo['karma']
        self.stable_karma = updateinfo['stable_karma']
        self.status = updateinfo['status']
        self.request = updateinfo['request']
        self.date_submitted = updateinfo['date_submitted']
        if updateinfo['date_pushed_testing']:
            self.date_pushed_testing = updateinfo['date_pushed_testing']
        if updateinfo['date_pushed_stable']:
            self.date_pushed_stable = updateinfo['date_pushed_stable']
        self.pending = updateinfo['pending']
        current_bugids = [bug.bugid for bug in self.bugs]
        for bugid in updateinfo['bugs']:
            # if full sync is specified, we want to sync all of the bugs
            if fullsync or not bugid in current_bugids:
                newbugs = Bug.query.filter_by(bugid=bugid).all()
                for bug in newbugs:
                    if not bugid in current_bugids:
                        self.bugs.append(bug)
                    if bug.milestone and not bug.milestone in self.milestones:
                        self.milestones.append(bug.milestone)

    @classmethod
    def from_data(cls, updateinfo, release):
        newupdate = Update(updateinfo['title'], '', 0, '', [], release, None)
        newupdate.sync(updateinfo)
        return newupdate
