#!/usr/bin/python3

import itertools
import datetime
import re

from blockerbugs import app
from blockerbugs.util import pagure_interface

TRACKER_KEYWORDS = [
    'betablocker',
    'finalblocker',
    'betafe',
    'finalfe',
    '0day',
    'previousrelease'
]


VOTE_KEYWORDS = [' '.join(v) for v in
                 itertools.product(TRACKER_KEYWORDS, ['-1', '0', '+1'])]

OUTCOME_KEYWORDS = [''.join(r) for r in
                   itertools.product(['accepted', 'rejected'], TRACKER_KEYWORDS)]

REVOTE_KEYWORDS = ['revote ' + t for t in TRACKER_KEYWORDS]

#ADMIN_KEYWORDS = RESULT_KEYWORDS + REVOTE_KEYWORDS

SUMMARY_HEADER = ""
CLOSED_VOTES_HEADER = "The following votes have been closed:"


def agreed_parser(line):
    out = []
    outcome_pattern = '|'.join(OUTCOME_KEYWORDS)
    #pattern = '^agreed (({}) )+.*$'.format(outcome_pattern)
    pattern = '^agreed ({})( ({}))*( .*)?$'.format(outcome_pattern, outcome_pattern)
    m = re.match(pattern, line)
    if m:
        for outcome in OUTCOME_KEYWORDS:
            if outcome in line:
                out.append('agreed ' + outcome)
    return out

def revote_parser(line):
    out = []
    trackers = '|'.join(TRACKER_KEYWORDS)
    pattern = '^revote (({}) ?)+$'.format(trackers)
    m = re.match(pattern, line)
    if m:
        for tracker in TRACKER_KEYWORDS:
            if tracker in line:
                out.append('revote ' + tracker)
    return out


def bot_loop_detected(comments):
    is_bot_comment = [comment.user == app.config['PAGURE_BOT_USERNAME'] for comment in comments]
    if len(is_bot_comment) < app.config['PAGURE_BOT_LOOP_THRESHOLD']:
        return False

    return all(is_bot_comment[-app.config['PAGURE_BOT_LOOP_THRESHOLD']:])


class Comment():
    def __init__(self, raw_comment):
        self.text = raw_comment['comment'].replace('\r', '')
        self.user = raw_comment['user']['name']
        self.id = raw_comment['id']

    def __repr__(self):
        return "%s (%s): %s" % (self.user, self.id, self.text)

    def user_is_admin(self):
        # FIXME use config for group name
        # FIXME it would be good to save/cache the data
        return self.user in pagure_interface.get_group('fedora-qa')['members']

    def keywords(self):
        out = []
        for line in self.text.lower().split('\n'):
            line = line.rstrip()
            line = re.sub(' +', ' ', line)
            if line in VOTE_KEYWORDS:
                out.append(line)
            elif line.startswith('agreed'):
                out.extend(agreed_parser(line))
            elif line.startswith('revote'):
                out.extend(revote_parser(line))
        return out

    def is_summary_post_of(self, tracker_keyword):
        is_bot = self.user == app.config['PAGURE_BOT_USERNAME']
        is_summary = CLOSED_VOTES_HEADER.lower() in self.text.lower()
        is_tracker_relevant = tracker_keyword in self.text.lower()

        return is_bot and is_summary and is_tracker_relevant


class BugVoteTracker():
    def __init__(self, accepted_keyword):
        self.open = True
        self.outcome = None
        self.accepted_keyword = accepted_keyword
        self.need_summary_post = False

        # vote format: {
        #     'user': {
        #          'vote': '+1'
        #          'comment_id': '21383'
        #     }
        # }
        self.votes = {}

    def _filter_relevant(self, keywords):
        accepted_votes = [self.accepted_keyword + v for v in [' +1', ' 0', ' -1']]
        accepted_agreed = ['agreed ' + o + self.accepted_keyword for o in ['accepted', 'rejected']]
        accepted_revote = ['revote ' + self.accepted_keyword]
        accepted_keywords = accepted_votes + accepted_agreed + accepted_revote
        return [keyword for keyword in keywords if keyword in accepted_keywords]

    def parse_comment(self, comment):
        if self.need_summary_post:
            self.need_summary_post = not comment.is_summary_post_of(self.accepted_keyword)

        keywords = self._filter_relevant(comment.keywords())
        for keyword in keywords:
            # FIXME log rejected (non admin) users?
            if keyword.startswith('agreed') and comment.user_is_admin():
                self.open = False
                self.outcome = keyword[7:15]  # eh...
                self.need_summary_post = True
            elif keyword.startswith('revote') and comment.user_is_admin():
                self.open = True
                self.votes = {}
            else:
                if not self.open:
                    continue
                for vote in ['-1', '0', '+1']:
                    if keyword.endswith(vote):
                        self.votes[comment.user] = {
                            'vote': vote,
                            'comment_id': comment.id
                        }
        return keywords

    def enumerate_votes(self):
        out = {
            '-1': [],
            '0': [],
            '+1': []
        }
        for user, vote in self.votes.items():
            out[vote['vote']].append((user, vote['comment_id']))

        return out

def link(user, comment_id):
    return "[%s](#comment-%s)" % (user, comment_id)

def tracker_summary(tracker_name, tracker):
    nicer = {
        'betablocker': 'BetaBlocker',
        'finalblocker': 'FinalBlocker',
        'betafe': 'BetaFE',
        'finalfe': 'FinalFE',
        '0day': '0Day',
        'previousrelease': 'PreviousRelease',
        'accepted': 'Accepted',
        'rejected': 'Rejected'
    }

    out = "* "

    outcome = ""
    if not tracker.open:
        outcome = "**%s** " % nicer[tracker.outcome]

    out += "%s%s " % (outcome, nicer[tracker_name])

    votes = tracker.enumerate_votes()
    pros = len(votes['+1'])
    neutrals = len(votes['0'])
    cons = len(votes['-1'])

    pros_text = "+%s" % pros
    if pros > 0:
        pros_text = "**+%s**" % pros

    cons_text = "-%s" % cons
    if cons > 0:
        cons_text = "**-%s**" % cons

    out += "(%s, %s, %s)\n" % (pros_text, neutrals, cons_text)

    return out

def summary(trackers, non_voting_users=None, last_comment_id=None, header=CLOSED_VOTES_HEADER):
    md_text = "%s\n\n" % header

    for tracker_name, tracker in trackers.items():
        if not tracker.votes:
            if not tracker.open:
                md_text += tracker_summary(tracker_name, tracker)
            continue

        votes = tracker.enumerate_votes()

        md_text += tracker_summary(tracker_name, tracker)

        for neco in ['+1', '0', '-1']:
            votes[neco] = [link(user, comment_id) for (user, comment_id) in votes[neco]]
            if len(votes[neco]) == 0:
                continue
            md_text += "    * %s by %s\n" % (neco, ', '.join(votes[neco]))

    md_text += '\n'

    if non_voting_users:
        md_text += "Commented but haven't voted yet: %s\n\n" % ', '.join([link(u, c) for (u, c) in non_voting_users.items()])

    if last_comment_id:
        time = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M")

        md_text += ("*The votes have been last counted at %s UTC and the last "
                    "processed comment was [#comment-%s](#comment-%s)*" % (time, last_comment_id, last_comment_id))

    return md_text


def parse_comments_to_trackers(comments):
    trackers = {tracker: BugVoteTracker(tracker) for tracker in TRACKER_KEYWORDS}

    for comment in comments:
        for tracker in trackers.values():
            tracker.parse_comment(comment)

    return trackers


def voting_info(comments, trackers):
    last_comment_id = 0
    non_voting_users = {}
    voting_users = set()

    for comment in comments:
        last_comment_id = comment.id
        if comment.user != app.config['PAGURE_BOT_USERNAME']:
            non_voting_users[comment.user] = comment.id

    for tracker in trackers.values():
        voting_users |= tracker.votes.keys()

    for user in voting_users:
        non_voting_users.pop(user)

    return non_voting_users, last_comment_id


def webhook_handler(issue_id):
    comments = [Comment(c) for c in pagure_interface.get_issue_comments(issue_id)]

    trackers = parse_comments_to_trackers(comments)

    non_voting_users, last_comment_id = voting_info(comments, trackers)

    trackers_without_summary = {n: t for (n, t) in trackers.items() if t.need_summary_post}
    if trackers_without_summary:
        app.logger.debug('Agreed/Rejected keyword found, posting summary comment')
        if bot_loop_detected(comments):
            app.logger.error('Bot is probably stuck in loop, not posting summary comment!')
        else:
            comment = summary(trackers_without_summary)
            pagure_interface.post_comment(issue_id, comment)

    app.logger.debug('Updating issue cummary')
    vote_summary = summary(trackers, non_voting_users, last_comment_id, SUMMARY_HEADER)
    pagure_interface.update_issue(issue_id, vote_summary)
